import os
import typer
import subprocess

from typing_extensions import Annotated
from clip_api_service import list_models as _list_models
from clip_api_service.build import build_bento
from clip_api_service.models import MODEL_ENV_VAR_KEY, DEFAULT_MODEL_NAME
app = typer.Typer()

@app.command()
def serve(model_name: Annotated[str, typer.Option(help="CLIP Model name")] = DEFAULT_MODEL_NAME):
    env = os.environ.copy()
    env[MODEL_ENV_VAR_KEY] = model_name

    try:
        subprocess.run(["bentoml", "serve", "clip_api_service.service:svc"], env=env, check=True)
    except subprocess.CalledProcessError as e:
        typer.echo(f"Command 'bentoml serve {model_name}' failed with error code {e.returncode}")


@app.command()
def build(
    model_name: Annotated[str, typer.Option(help="CLIP Model name")] = DEFAULT_MODEL_NAME,
    use_gpu: Annotated[bool, typer.Option(help="Use GPU for build")] = False,
):
    build_bento(model_name=model_name, use_gpu=use_gpu)

@app.command()
def list_models():
    print(_list_models())

if __name__ == "__main__":
    app()


from __future__ import annotations

import base64
import io
from typing import Any, List

import aiohttp
import numpy as np
from bentoml.exceptions import BadInput
from PIL import Image
from pydantic import BaseModel, RootModel


def base64_to_image(base64_string: str) -> Image:
    image_data = base64.b64decode(base64_string)
    image_io = io.BytesIO(image_data)
    return Image.open(image_io)


async def download_image_from_url(url: str) -> Image.Image:
    async with aiohttp.ClientSession() as session:
        try:
            async with session.get(url, raise_for_status=True) as response:
                image_data = await response.read()
                image = Image.open(io.BytesIO(image_data)).convert("RGB")
                return image
        except aiohttp.client_exceptions.ClientResponseError as ex:
            raise BadInput(f"Failed to download image from {url}") from ex


def cosine_similarity(query_embeds, candidates_embeds):
    # Normalize each embedding to a unit vector
    query_embeds /= np.linalg.norm(query_embeds, axis=1, keepdims=True)
    candidates_embeds /= np.linalg.norm(candidates_embeds, axis=1, keepdims=True)

    # Compute cosine similarity
    cosine_similarities = np.matmul(query_embeds, candidates_embeds.T)

    return cosine_similarities


def softmax(scores):
    # Compute softmax scores (probabilities)
    exp_scores = np.exp(
        scores - np.max(scores, axis=-1, keepdims=True)
    )  # Subtract max for numerical stability
    return exp_scores / np.sum(exp_scores, axis=-1, keepdims=True)


class BaseItem(BaseModel):
    def dict(self, *args, **kwargs) -> dict[str, Any]:
        """
        Override the default dict method to exclude None values in the response
        """
        kwargs.pop("exclude_none", None)
        return super().dict(*args, exclude_none=True, **kwargs)


class ListModel(RootModel):
    root: List[Any]

    def __iter__(self):
        return iter(self.root)

    def __getitem__(self, item):
        return self.root[item]

    def append(self, item):
        return self.root.append(item)



from __future__ import annotations

import bentoml

from clip_api_service.models import get_clip_module, init_model


def get_clip_runner(model: str | None | bentoml.Model = None, init_local: bool = False):
    bento_model = model if isinstance(model, bentoml.Model) else init_model(model)
    clip_module = get_clip_module(bento_model)

    runner = bentoml.Runner(
        clip_module.clip_runnable(),
        name="clip_model_runner",
        runnable_init_params={"bento_model": bento_model},
        models=[bento_model],
    )

    if init_local:
        runner.init_local(quiet=True)

    return runner



from __future__ import annotations

import typing

import bentoml
import torch
from PIL import Image

from clip_api_service.models import CLIPRunnable

if typing.TYPE_CHECKING:
    import numpy.typing as npt

MODELS = {
    "RN50:openai",
    "RN50:yfcc15m",
    "RN50:cc12m",
    # "RN50-quickgelu:openai",
    # "RN50-quickgelu:yfcc15m",
    # "RN50-quickgelu:cc12m",
    "RN101:openai",
    "RN101:yfcc15m",
    # "RN101-quickgelu:openai",
    # "RN101-quickgelu:yfcc15m",
    "RN50x4:openai",
    "RN50x16:openai",
    "RN50x64:openai",
    "ViT-B-32:openai",
    "ViT-B-32:laion400m_e31",
    "ViT-B-32:laion400m_e32",
    "ViT-B-32:laion2b_e16",
    "ViT-B-32:laion2b_s34b_b79k",
    # "ViT-B-32:datacomp_m_s128m_b4k",
    # "ViT-B-32:commonpool_m_clip_s128m_b4k",
    # "ViT-B-32:commonpool_m_laion_s128m_b4k",
    # "ViT-B-32:commonpool_m_image_s128m_b4k",
    # "ViT-B-32:commonpool_m_text_s128m_b4k",
    # "ViT-B-32:commonpool_m_basic_s128m_b4k",
    # "ViT-B-32:commonpool_m_s128m_b4k",
    # "ViT-B-32:datacomp_s_s13m_b4k",
    # "ViT-B-32:commonpool_s_clip_s13m_b4k",
    # "ViT-B-32:commonpool_s_laion_s13m_b4k",
    # "ViT-B-32:commonpool_s_image_s13m_b4k",
    # "ViT-B-32:commonpool_s_text_s13m_b4k",
    # "ViT-B-32:commonpool_s_basic_s13m_b4k",
    # "ViT-B-32:commonpool_s_s13m_b4k",
    # "ViT-B-32-quickgelu:openai",
    # "ViT-B-32-quickgelu:laion400m_e31",
    # "ViT-B-32-quickgelu:laion400m_e32",
    "ViT-B-16:openai",
    "ViT-B-16:laion400m_e31",
    "ViT-B-16:laion400m_e32",
    # "ViT-B-16:laion2b_s34b_b88k",
    # "ViT-B-16:datacomp_l_s1b_b8k",
    # "ViT-B-16:commonpool_l_clip_s1b_b8k",
    # "ViT-B-16:commonpool_l_laion_s1b_b8k",
    # "ViT-B-16:commonpool_l_image_s1b_b8k",
    # "ViT-B-16:commonpool_l_text_s1b_b8k",
    # "ViT-B-16:commonpool_l_basic_s1b_b8k",
    # "ViT-B-16:commonpool_l_s1b_b8k",
    "ViT-B-16-plus-240:laion400m_e31",
    "ViT-B-16-plus-240:laion400m_e32",
    "ViT-L-14:openai",
    "ViT-L-14:laion400m_e31",
    "ViT-L-14:laion400m_e32",
    "ViT-L-14:laion2b_s32b_b82k",
    # "ViT-L-14:datacomp_xl_s13b_b90k",
    # "ViT-L-14:commonpool_xl_clip_s13b_b90k",
    # "ViT-L-14:commonpool_xl_laion_s13b_b90k",
    # "ViT-L-14:commonpool_xl_s13b_b90k",
    "ViT-L-14-336:openai",
    "ViT-H-14:laion2b_s32b_b79k",
    "ViT-g-14:laion2b_s12b_b42k",
    "ViT-g-14:laion2b_s34b_b88k",
    "ViT-bigG-14:laion2b_s39b_b160k",
    "roberta-ViT-B-32:laion2b_s12b_b32k",
    "xlm-roberta-base-ViT-B-32:laion5b_s13b_b90k",
    "xlm-roberta-large-ViT-H-14:frozen_laion5b_s13b_b90k",
    # "convnext_base:laion400m_s13b_b51k",
    # "convnext_base_w:laion2b_s13b_b82k",
    # "convnext_base_w:laion2b_s13b_b82k_augreg",
    # "convnext_base_w:laion_aesthetic_s13b_b82k",
    # "convnext_base_w_320:laion_aesthetic_s13b_b82k",
    # "convnext_base_w_320:laion_aesthetic_s13b_b82k_augreg",
    # "convnext_large_d:laion2b_s26b_b102k_augreg",
    # "convnext_large_d_320:laion2b_s29b_b131k_ft",
    # "convnext_large_d_320:laion2b_s29b_b131k_ft_soup",
    # "convnext_xxlarge:laion2b_s34b_b82k_augreg",
    # "convnext_xxlarge:laion2b_s34b_b82k_augreg_rewind",
    # "convnext_xxlarge:laion2b_s34b_b82k_augreg_soup",
    # "coca_ViT-B-32:laion2b_s13b_b90k",
    # "coca_ViT-B-32:mscoco_finetuned_laion2b_s13b_b90k",
    # "coca_ViT-L-14:laion2b_s13b_b90k",
    # "coca_ViT-L-14:mscoco_finetuned_laion2b_s13b_b90k",
    # "EVA01-g-14:laion400m_s11b_b41k",
    # "EVA01-g-14-plus:merged2b_s11b_b114k",
    # "EVA02-B-16:merged2b_s8b_b131k",
    # "EVA02-L-14:merged2b_s4b_b131k",
    # "EVA02-L-14-336:merged2b_s6b_b61k",
    # "EVA02-E-14:laion2b_s4b_b115k",
    # "EVA02-E-14-plus:laion2b_s9b_b144k",
}



def get_bento_model_tag(model_name: str) -> bentoml.Tag:
    model_version = model_name.replace(":", ".")
    return bentoml.Tag("openclip", model_version.lower())


def download_model(model_name: str) -> bentoml.Model:
    import open_clip

    bento_model_tag = get_bento_model_tag(model_name)
    _model_name, _pretrained = model_name.split(":")

    model, _, processor = open_clip.create_model_and_transforms(
        _model_name, pretrained=_pretrained
    )
    tokenizer = open_clip.get_tokenizer(_model_name)

    bentoml.pytorch.save_model(
        str(bento_model_tag),
        model,
        custom_objects={"processor": processor, "tokenizer": tokenizer},
        labels={
            "clip_module": "clip_api_service.models.openclip",
        },
        metadata={
            "logit_scale": model.logit_scale.item(),
        },
    )
    return bentoml.models.get(bento_model_tag)

def bentofile_path(use_gpu: bool = False) -> str:
    import os
    build_ctx = os.path.dirname(os.path.dirname(__file__))

    declaration = {
        False: os.path.join(build_ctx, "bentofiles", "bentofile.openclip.cpu.yaml"),
        True: os.path.join(build_ctx, "bentofiles", "bentofile.openclip.gpu.yaml"),
    }

    return declaration[use_gpu]

class OpenClipRunnable(CLIPRunnable):
    SUPPORTED_RESOURCES = ("nvidia.com/gpu", "cpu")
    SUPPORTS_CPU_MULTI_THREADING = True

    def __init__(self, bento_model: bentoml.Model):
        import torch

        if torch.cuda.is_available():
            self.device = "cuda"
            # by default, torch.FloatTensor will be used on CPU.
            torch.set_default_tensor_type("torch.cuda.FloatTensor")
        else:
            self.device = "cpu"

        self.model = bento_model.load_model().to(self.device).eval()
        self.processor = bento_model.custom_objects["processor"]
        self.tokenizer = bento_model.custom_objects["tokenizer"]

    @bentoml.Runnable.method(batchable=True)
    def encode_text(self, texts: list[str]) -> npt.NDArray:
        texts_encodings = self.tokenizer(texts)
        with torch.inference_mode():
            text_embeddings = self.model.encode_text(texts_encodings.to(self.device))
            return text_embeddings.cpu().detach().numpy()

    @bentoml.Runnable.method(batchable=True)
    def encode_image(self, images: list[Image.Image]) -> npt.NDArray:
        image_encodings = torch.stack([self.processor(image) for image in images])
        with torch.inference_mode():
            image_embeddings = self.model.encode_image(image_encodings.to(self.device))
            return image_embeddings.cpu().detach().numpy()


def clip_runnable():
    return OpenClipRunnable



from __future__ import annotations

import importlib
import logging
import os
import typing
from typing import Protocol, cast

import bentoml
from bentoml.exceptions import NotFound
from PIL import Image

if typing.TYPE_CHECKING:
    from typing import Any

    import numpy.typing as npt

MODEL_ENV_VAR_KEY = "CLIP_MODEL"
DEFAULT_MODEL_NAME = "openai/clip-vit-large-patch14"

logger = logging.getLogger(__name__)


class CLIPRunnable(bentoml.Runnable):
    """
    CLIPRunnable is a protocol for defining a runnable that implements the actual
    CLIP encoding logic. The interface is used by the CLIP API service to encode text
    and images into CLIP embeddings. See service.py for more details.
    """

    def __init__(self, bento_model: bentoml.Model):
        ...

    def encode_text(self, texts: list[str]) -> npt.NDArray:
        ...

    def encode_image(self, images: list[Image.Image]) -> npt.NDArray:
        ...


class CLIPModule(Protocol):
    """
    CLIPModule is a protocol for defining a CLIP flavor. A CLIP module is a
    Python module that contains a set of CLIP models. Each CLIP model is
    identified by a string name. The CLIP module is responsible for saving
    and loading the CLIP models. It also provides a CLIPRunnable class that
    implements the actual CLIP encoding logic for serving with BentoML.
    """

    MODELS: set[str]

    def download_model(
        self, model_name: str, model: Any, custom_objects: dict[str, Any]
    ) -> bentoml.Model:
        ...

    def clip_runnable(self) -> type[CLIPRunnable]:
        # TODO: add *args, **kwargs for customizing Runnable behavior
        ...

    def bentofile_path(self, use_gpu: bool = False) -> str:
        ...

    def get_bento_model_tag(self, model_name) -> bentoml.Tag:
        ...


class CLIPModuleRegistry:
    def __init__(self):
        self._modules: list[CLIPModule] = []
        self._register_default_clip_modules()

    def register(self, module: str | CLIPModule) -> None:
        if isinstance(module, str):
            self._modules.append(cast(CLIPModule, importlib.import_module(module)))
        else:
            self._modules.append(module)

    def _register_default_clip_modules(self):
        from clip_api_service.models import openai, openclip  # cclip, mclip

        self.register(cast(CLIPModule, openai))
        self.register(cast(CLIPModule, openclip))
        # self.register(cast(CLIPModule, cclip))
        # self.register(cast(CLIPModule, mclip))

    def __iter__(self):
        return iter(self._modules)


CLIP_MODULES = CLIPModuleRegistry()


def save_model(
    model_name: str,
    clip_module: str,
    model: Any,
    custom_objects: dict[str, Any],
) -> bentoml.Model:
    """
    Save a CLIP model to BentoML format. A clip_module must be specified to indicate
    its implementation. The clip_module must be a python module path that defines the
    CLIPModule protocol. A CLIPModule defines the following methods:
        - download_model(model_name: str, model: Any, custom_objects: Dict[str, Any]) -> bentoml.Model
        - clip_runnable() -> type[CLIPRunnable]
        - bentofile_path(use_gpu: bool = False) -> str
        - get_bento_model_tag(model_name) -> bentoml.Tag

    See clip_api_service.models.openai for an example.
    """
    if clip_module not in [
        "clip_api_service.models.openai",
        "clip_api_service.models.openclip",
        "clip_api_service.models.cclip",
        "clip_api_service.models.mclip",
    ]:
        logger.info("Using custom clip module %s", clip_module)

    _clip_module = importlib.import_module(clip_module)
    return _clip_module.save_model(
        model_name, model, custom_objects, labels={"clip_module": clip_module}
    )


def init_model(model_name: str | None = None) -> bentoml.Model:
    """
    Initialize a CLIP model from BentoML format. The model_name must be specified to
    support multiple models in a single BentoML bundle.
    """
    if model_name is None:
        model_name = os.environ.get(MODEL_ENV_VAR_KEY, DEFAULT_MODEL_NAME)

    for clip_module in CLIP_MODULES:
        if model_name in clip_module.MODELS:
            model_tag = clip_module.get_bento_model_tag(model_name)
            try:
                return bentoml.models.get(model_tag)
            except bentoml.exceptions.NotFound:
                return clip_module.download_model(model_name)

    raise NotFound(f"Model {model_name} not supported in any registered clip module")


def get_clip_module(bento_model: bentoml.Model) -> CLIPModule:
    clip_module_name = bento_model.info.labels.get("clip_module", None)
    if clip_module_name is None:
        raise ValueError("clip_module not found in model labels")

    clip_module = cast(CLIPModule, importlib.import_module(clip_module_name))
    assert hasattr(
        clip_module, "clip_runnable"
    ), f'clip_runnable not found in clip_module "{clip_module_name}"'
    return clip_module


def list_models() -> list[str]:
    models = []
    for clip_module in CLIP_MODULES:
        models.extend(clip_module.MODELS)
    return models



from __future__ import annotations

import asyncio
import typing

import bentoml
import numpy as np
from bentoml.io import JSON
from typing import Optional, List

from clip_api_service.models import init_model
from clip_api_service.runners import get_clip_runner
from clip_api_service.samples import ENCODING_INPUT_SAMPLE, RANKING_INPUT_SAMPLE
from clip_api_service.utils import (
    BaseItem,
    ListModel,
    base64_to_image,
    cosine_similarity,
    download_image_from_url,
    softmax,
)

if typing.TYPE_CHECKING:
    import numpy.typing as npt


class Item(BaseItem):
    text: Optional[str] = None 
    img_uri: Optional[str] = None  
    img_blob: Optional[str] = None 


class ItemList(ListModel):
    root: List[Item]


class RankInput(BaseItem):
    queries: List[Item]
    candidates: List[Item]


class RankOutput(BaseItem):
    probabilities: List[List[float]]
    cosine_similarities: List[List[float]]

bento_model = init_model("__model_name__")
logit_scale = np.exp(bento_model.info.metadata.get("logit_scale", 4.60517))

clip_runner = get_clip_runner(bento_model)
svc = bentoml.Service(
    "clip-api-service",
    runners=[clip_runner],
)


async def _encode(item: Item) -> npt.NDArray:
    if item.img_uri:
        image = await download_image_from_url(item.img_uri)
        embedding = await clip_runner.encode_image.async_run([image])
    elif item.img_blob:
        image = base64_to_image(item.img_blob)
        embedding = await clip_runner.encode_image.async_run([image])
    else:
        embedding = await clip_runner.encode_text.async_run([item.text])

    return embedding[0]


@svc.api(
    input=JSON.from_sample(ENCODING_INPUT_SAMPLE, pydantic_model=ItemList),
    output=JSON(),
)
async def encode(items: ItemList) -> List[npt.NDArray[float]]:
    results = [_encode(item) for item in items]
    results = await asyncio.gather(*results)
    return results


@svc.api(
    input=JSON.from_sample(RANKING_INPUT_SAMPLE, pydantic_model=RankInput),
    output=JSON(pydantic_model=RankOutput),
)
async def rank(rank_input: RankInput) -> RankOutput:
    queries = [_encode(query) for query in rank_input.queries]
    candidates = [_encode(item) for item in rank_input.candidates]

    # Encode embeddings
    query_embeds = np.array(await asyncio.gather(*queries))
    candidate_embeds = np.array(await asyncio.gather(*candidates))

    # Compute cosine similarities
    cosine_similarities = cosine_similarity(query_embeds, candidate_embeds)

    # Compute softmax scores
    prob_scores = softmax(logit_scale * cosine_similarities)
    return RankOutput(
        probabilities=prob_scores.tolist(),
        cosine_similarities=cosine_similarities.tolist(),
    )



from __future__ import annotations

import typing

import bentoml
from PIL import Image

from clip_api_service.models import CLIPRunnable

if typing.TYPE_CHECKING:
    import numpy.typing as npt

MODELS = {
    "openai/clip-vit-large-patch14",
    "openai/clip-vit-base-patch32",
    "openai/clip-vit-base-patch16",
    "openai/clip-vit-large-patch14-336",
}


def get_bento_model_tag(model_name: str) -> bentoml.Tag:
    return bentoml.Tag.from_str(model_name.replace("openai/clip-", "openai-clip:"))


def download_model(model_name: str) -> bentoml.Model:
    from transformers import CLIPModel, CLIPProcessor

    model = CLIPModel.from_pretrained(model_name)
    processor = CLIPProcessor.from_pretrained(model_name)
    bento_model_tag = get_bento_model_tag(model_name)

    bentoml.transformers.save_model(
        str(bento_model_tag),
        model,
        custom_objects={"processor": processor},
        labels={
            "clip_module": "clip_api_service.models.openai",
        },
        metadata={
            "logit_scale": model.logit_scale.item(),
        },
    )
    return bentoml.models.get(bento_model_tag)

def bentofile_path(use_gpu: bool = False) -> str:
    import os
    build_ctx = os.path.dirname(os.path.dirname(__file__))

    declaration = {
        False: os.path.join(build_ctx, "bentofiles", "bentofile.openai.cpu.yaml"),
        True: os.path.join(build_ctx, "bentofiles", "bentofile.openai.gpu.yaml"),
    }

    return declaration[use_gpu]

class OpenAICLIPRunnable(CLIPRunnable):
    SUPPORTED_RESOURCES = ("nvidia.com/gpu", "cpu")
    SUPPORTS_CPU_MULTI_THREADING = True

    def __init__(self, bento_model: bentoml.Model):
        import torch

        if torch.cuda.is_available():
            self.device = "cuda"
            # by default, torch.FloatTensor will be used on CPU.
            torch.set_default_tensor_type("torch.cuda.FloatTensor")
        else:
            self.device = "cpu"

        self.model = bento_model.load_model().to(self.device)
        self.processor = bento_model.custom_objects["processor"]

    @bentoml.Runnable.method(batchable=True)
    def encode_text(self, texts: list[str]) -> npt.NDArray:
        inputs = self.processor(text=texts, return_tensors="pt", padding=True).to(
            self.device
        )
        text_embeddings = self.model.get_text_features(**inputs)
        return text_embeddings.cpu().detach().numpy()

    @bentoml.Runnable.method(batchable=True)
    def encode_image(self, images: list[Image.Image]) -> npt.NDArray:
        inputs = self.processor(images=images, return_tensors="pt", padding=True).to(
            self.device
        )
        image_embeddings = self.model.get_image_features(**inputs)
        return image_embeddings.cpu().detach().numpy()


def clip_runnable():
    return OpenAICLIPRunnable



from __future__ import annotations

# Chinese CLIP models
MODELS = {
    "CN-CLIP/ViT-B-16",
    "CN-CLIP/ViT-L-14",
    "CN-CLIP/ViT-L-14-336",
    "CN-CLIP/ViT-H-14",
    "CN-CLIP/RN50",
}

raise NotImplementedError("CNCLIP is not implemented yet.")



from __future__ import annotations

SAMPLE_IMG_URL = "https://hips.hearstapps.com/hmg-prod/images/dog-puppy-on-garden-royalty-free-image-1586966191.jpg"

ENCODING_INPUT_SAMPLE = [
    {"img_uri": SAMPLE_IMG_URL},
    {"text": "picture of a dog"},
    {"text": "picture of a cat"},
    {"text": "picture of a bird"},
    {"text": "picture of a car"},
    {"text": "picture of a plane"},
    {"text": "picture of a boat"},
]

RANKING_INPUT_SAMPLE = {
    "queries": [
        {"img_uri": SAMPLE_IMG_URL},
    ],
    "candidates": [
        {"text": "picture of a dog"},
        {"text": "picture of a cat"},
        {"text": "picture of a bird"},
        {"text": "picture of a car"},
        {"text": "picture of a plane"},
        {"text": "picture of a boat"},
    ],
}



from __future__ import annotations

# Multi-lingual CLIP models
MODELS = {
    "M-CLIP/XLM-Roberta-Large-Vit-B-16Plus",
    "M-CLIP/M-BERT-Distil-40",
    "M-CLIP/M-BERT-Base-ViT-B",
    "M-CLIP/XLM-Roberta-Large-Vit-B-32",
    "M-CLIP/XLM-Roberta-Large-Vit-L-14",
}

raise NotImplementedError("MCCLIP is not implemented yet.")



from __future__ import annotations

from clip_api_service.models import (
    CLIP_MODULES as modules_registry,
)
from clip_api_service.models import (
    init_model,
    list_models,
    save_model,
)
from clip_api_service.runners import get_clip_runner

# def start_clip_server(in_process: bool=False) -> tuple[bentoml.HTTPServer, bentoml.client.HTTPClient]:
#     from clip_api_service.service import svc
#
#     if in_process:
#         pass
#     else:
#         server = bentoml.HTTPServer(svc)
#         server.start(blocking=False)
#         client = server.get_client()
#         return server, client


__all__ = [
    "get_clip_runner",
    "save_model",
    "init_model",
    "modules_registry",
    "list_models",
]



from __future__ import annotations

from clip_api_service.models import MODEL_ENV_VAR_KEY

import os


def build_bento(model_name: str | None = None, use_gpu: bool = False):
    import bentoml.bentos

    from clip_api_service.models import get_clip_module, init_model

    bento_model = init_model(model_name)
    clip_module = get_clip_module(bento_model)

    bento_file = clip_module.bentofile_path(use_gpu=use_gpu)

    # Setting Model name as environment variable to pass 
    # into subprocess for bento build
    if model_name:
        os.environ[MODEL_ENV_VAR_KEY] = model_name

    # Generate a service file
    # TODO: Refactor the code gen to a mode generalize function
    # TODO: Use temporary file instead of writing to the source file 
    src_service_file_path = os.path.join(os.path.dirname(__file__), "_service.py")
    target_service_file_path = os.path.join(os.path.dirname(__file__), "service.py")

    with open(src_service_file_path, 'r') as src_file, open(target_service_file_path, 'w') as target_file:
        data = src_file.read().replace('__model_name__', model_name)
        target_file.write(data)

    # Get parent directory path
    build_ctx = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    bentoml.bentos.build_bentofile(bento_file, build_ctx=build_ctx)

    # Remove the generated service file
    os.remove(target_service_file_path)